""" Tests the dictionary based FC evaluator"""
from unittest import mock

import pytest

from ahbicht.content_evaluation.fc_evaluators import DictBasedFcEvaluator
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint

pytestmark = pytest.mark.asyncio


class TestDictBasedFcEvaluator:
    """Test for the evaluation using the Dict Based FC Evaluator"""

    async def test_evaluation(self, mocker):
        hardcoded_results = {
            "1": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
            "2": EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something wrong"),
        }
        evaluator = DictBasedFcEvaluator(hardcoded_results)
        assert await evaluator.evaluate_single_format_constraint("1", entered_input="asd") == EvaluatedFormatConstraint(
            format_constraint_fulfilled=True
        )
        assert await evaluator.evaluate_single_format_constraint("2", entered_input="yxc") == EvaluatedFormatConstraint(
            format_constraint_fulfilled=False, error_message="something wrong"
        )
        with pytest.raises(NotImplementedError):
            await evaluator.evaluate_single_format_constraint("3", entered_input="qwe")
        dict_evaluation_spy = mocker.spy(evaluator, "evaluate_single_format_constraint")
        assert await evaluator.evaluate_format_constraints(["1", "2"], entered_input="asd") == hardcoded_results
        dict_evaluation_spy.assert_has_awaits([mock.call("1", "asd"), mock.call("2", "asd")])



""" Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung) """

from unittest.mock import AsyncMock

import inject
import pytest

from ahbicht.condition_check_results import FormatConstraintEvaluationResult, RequirementConstraintEvaluationResult
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.hints_provider import HintsProvider


class TestAHBExpressionEvaluation:
    """Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung)"""

    @pytest.fixture()
    def setup_and_teardown_injector(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(HintsProvider, AsyncMock(wraps=HintsProvider)).bind(
                RcEvaluator, AsyncMock(wraps=RcEvaluator)
            )
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        """ahb_expression, expected_requirement_indicator, expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional, expected_format_constraints_expression, expected_hints""",
        [
            pytest.param("Muss", "Muss", True, False, None, None),
            pytest.param("X", "X", True, False, None, None),
            pytest.param("Muss[1]", "Muss", True, True, None, None),
            pytest.param("Muss[2]", "Muss", False, True, None, None),
            pytest.param("Muss[1]Soll[2]", "Muss", True, True, None, None),
            pytest.param("Muss[2]Soll[1]Kann[4]", "Soll", True, True, None, None),
            pytest.param("Muss[2]\tSoll [ 1]  Kann[4\t]", "Soll", True, True, None, None),
            pytest.param("X[1]", "X", True, True, None, None),
            pytest.param("O[2]", "O", False, True, None, None),
            pytest.param("U[1]", "U", True, True, None, None),
            pytest.param("U[1]U[2]", "U", False, True, None, None),
            pytest.param("X[1]U[2]U[3]O[4]", "X", False, True, None, None),
            pytest.param("Muss([1]O[2])U[3]Soll[2]Kann[2]O[4]", "Muss", True, True, None, None),
            pytest.param("Muss[2]Soll[2]", "Soll", False, True, None, None),
            # Neutral value
            pytest.param("Muss[503]", "Muss", True, False, None, "[503]"),
            pytest.param("Soll[902]", "Soll", True, False, "[902]", None),
            pytest.param("Kann[503]U[504]", "Kann", True, False, None, "[503]U[504]"),
            # Last modal mark without conditions
            pytest.param("Muss[1]Kann", "Muss", True, True, None, None),
            pytest.param("Muss[2]Kann", "Kann", True, True, None, None),
        ],
    )
    def test_evaluate_valid_ahb_expression(
        self,
        mocker,
        ahb_expression,
        expected_requirement_indicator,
        expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional,
        expected_format_constraints_expression,
        expected_hints,
    ):
        """
        Tests that valid ahb expressions are evaluated as expected.
        Odd condition_keys are True, even condition_keys are False
        """

        def side_effect_rc_evaluation(condition_expression):
            if condition_expression in ["[1]", " [ 1]  ", "[3]", "([1]O[2])U[3]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression in ["[2]", "[2]\t", "[4]", "[4\t]", "[1]U[2]", "[1]U[2]U[3]O[4]", "[2]O[4]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=False,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression in ["[503]", "[503]U[504]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=None,
                    hints=condition_expression,
                )
            if condition_expression in ["[902]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=condition_expression,
                    hints=None,
                )

        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.requirement_constraint_evaluation",
            side_effect=side_effect_rc_evaluation,
        )
        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.format_constraint_evaluation",
            return_value=FormatConstraintEvaluationResult(format_constraints_fulfilled=True, error_message=None),
        )

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        result = evaluate_ahb_expression_tree(parsed_tree, entered_input=None)

        assert result.requirement_indicator == expected_requirement_indicator
        assert (
            result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            == expected_requirement_constraints_fulfilled
        )
        assert (
            result.requirement_constraint_evaluation_result.requirement_is_conditional
            == expected_requirement_is_conditional
        )
        assert (
            result.requirement_constraint_evaluation_result.format_constraints_expression
            == expected_format_constraints_expression
        )
        assert result.requirement_constraint_evaluation_result.hints == expected_hints

    @pytest.mark.parametrize(
        "expression, expected_error, expected_error_message",
        [
            pytest.param(
                "Muss[]",
                SyntaxError,
                """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """,
            ),
            pytest.param(
                "Soll[1001]",
                ValueError,
                """Condition key is not in valid number range.""",
            ),
        ],
    )
    def test_invalid_ahb_expressions(
        self, expression: str, expected_error: type, expected_error_message: str, setup_and_teardown_injector
    ):
        """Tests that an error is raised when trying to pass invalid values."""

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)

        with pytest.raises(expected_error) as excinfo:
            evaluate_ahb_expression_tree(parsed_tree, entered_input=None)

        assert expected_error_message in str(excinfo.value)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

from sphinx.ext import apidoc

project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, project_root)

# -- Project information -----------------------------------------------------

project = "ahbicht"
copyright = "2021, Annika Schloegl"
author = "Annika Schloegl"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",  # translate google/numpy docstring
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "./_static/ahbicht-logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "./_static/ahbicht-favicon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Auto-generate API documentation
# Source: https://github.com/readthedocs/readthedocs.org/issues/1139
def run_apidoc(_):
    output_path = os.path.join(project_root, "docs", "apidoc")
    module_path = os.path.join(project_root, "src/ahbicht")
    apidoc.main(["-o", output_path, module_path, "--separate", "--force"])


def setup(app):
    app.connect("builder-inited", run_apidoc)



""" Tests for Class Condition Node Builder"""
from pathlib import Path

import inject
import pytest

from ahbicht.condition_node_builder import ConditionNodeBuilder
from ahbicht.content_evaluation.evaluationdatatypes import EvaluationContext
from ahbicht.content_evaluation.rc_evaluators import EvaluatableData, RcEvaluator
from ahbicht.edifact import EdifactFormat, EdifactFormatVersion
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.hints_provider import HintsProvider, JsonFileHintsProvider


class DummyRcEvaluator(RcEvaluator):
    """
    a dummy requirement constraint evaluator to be used in the test
    """

    def _get_default_context(self) -> EvaluationContext:
        return None

    edifact_format = EdifactFormat.UTILMD
    edifact_format_version = EdifactFormatVersion.FV2104


class TestConditionNodeBuilder:
    _edifact_format = EdifactFormat.UTILMD
    _edifact_format_version = EdifactFormatVersion.FV2104

    _h_583 = Hint(condition_key="583", hint="[583] Hinweis: Verwendung der ID der Marktlokation")
    _h_584 = Hint(condition_key="584", hint="[584] Hinweis: Verwendung der ID der Messlokation")
    _ufc_955 = UnevaluatedFormatConstraint(condition_key="955")
    _ufc_907 = UnevaluatedFormatConstraint(condition_key="907")

    @pytest.fixture()
    def setup_and_teardown_injector(self):
        _hints_provider = JsonFileHintsProvider(
            TestConditionNodeBuilder._edifact_format,
            TestConditionNodeBuilder._edifact_format_version,
            file_path=Path("unittests/resources_condition_hints/FV2104/Hints_FV2104_UTILMD.json"),
        )
        inject.clear_and_configure(
            lambda binder: binder.bind(HintsProvider, _hints_provider).bind(
                RcEvaluator, DummyRcEvaluator(EvaluatableData(edifact_seed=dict()))
            )
        )
        yield
        inject.clear()

    def test_initiating_condition_node_builder(self, setup_and_teardown_injector):
        """Tests if condition node builder is initiated correctly."""

        condition_keys = ["501", "12", "903"]
        condition_node_builder = ConditionNodeBuilder(condition_keys)

        assert condition_node_builder.hints_provider.edifact_format == self._edifact_format
        assert condition_node_builder.rc_evaluator.edifact_format_version == self._edifact_format_version
        assert condition_node_builder.condition_keys == condition_keys
        assert condition_node_builder.requirement_constraints_condition_keys == ["12"]
        assert condition_node_builder.hints_condition_keys == ["501"]
        assert condition_node_builder.format_constraints_condition_keys == ["903"]

    def test_invalid_initiating_condition_node_builder(self, setup_and_teardown_injector):
        """Test that correct error is shown if condition keys are out of range."""
        condition_keys = ["5", "1011"]

        with pytest.raises(ValueError) as excinfo:
            _ = ConditionNodeBuilder(condition_keys)

        assert "Condition key is not in valid number range." in str(excinfo.value)

    def test_build_hint_nodes(self, setup_and_teardown_injector):
        """Tests that hint nodes are build correctly."""
        condition_keys = ["584", "583"]
        condition_node_builder = ConditionNodeBuilder(condition_keys)
        hint_nodes = condition_node_builder._build_hint_nodes()
        excepted_hints_nodes = {"583": self._h_583, "584": self._h_584}
        assert hint_nodes == excepted_hints_nodes

    def test_invalid_hint_nodes(self, setup_and_teardown_injector):
        """Tests that correct error is shown, when hint is not implemented."""
        condition_keys = ["500"]
        # it is possible that a hint with [500] will be implemented in the future as not all hints are collected yet.
        # If test fails, look up if hint exist. And if hint list is completed take one that does not exist.
        condition_node_builder = ConditionNodeBuilder(condition_keys)
        with pytest.raises(KeyError) as excinfo:
            _ = condition_node_builder._build_hint_nodes()

        assert "There seems to be no hint implemented with condition key '500'." in str(excinfo.value)

    def test_build_unevaluated_format_constraint_nodes(self, setup_and_teardown_injector):
        """Tests that unevaluated format constraints nodes are build correctly."""
        condition_keys = ["907", "955"]
        condition_node_builder = ConditionNodeBuilder(condition_keys)
        unevaluated_fc_nodes = condition_node_builder._build_unevaluated_format_constraint_nodes()
        expected_unevaluated_fc_nodes = {"907": self._ufc_907, "955": self._ufc_955}
        assert unevaluated_fc_nodes == expected_unevaluated_fc_nodes

    @pytest.mark.parametrize(
        "expected_conditions_fulfilled_11, expected_conditions_fulfilled_78",
        [
            pytest.param(ConditionFulfilledValue.FULFILLED, ConditionFulfilledValue.UNKNOWN),
            pytest.param(ConditionFulfilledValue.UNFULFILLED, ConditionFulfilledValue.NEUTRAL),
        ],
    )
    def test_build_requirement_constraint_nodes(
        self, mocker, expected_conditions_fulfilled_11, expected_conditions_fulfilled_78, setup_and_teardown_injector
    ):
        """Tests that requirement constraint nodes are build correctly."""

        mocker.patch(
            "ahbicht.content_evaluation.rc_evaluators.RcEvaluator.evaluate_single_condition",
            side_effect=[expected_conditions_fulfilled_11, expected_conditions_fulfilled_78],
        )

        condition_keys = ["11", "78"]
        condition_node_builder = ConditionNodeBuilder(condition_keys)

        evaluated_requirement_constraints = condition_node_builder._build_requirement_constraint_nodes()

        expected_requirement_constraints = {
            "11": RequirementConstraint(condition_key="11", conditions_fulfilled=expected_conditions_fulfilled_11),
            "78": RequirementConstraint(condition_key="78", conditions_fulfilled=expected_conditions_fulfilled_78),
        }
        assert evaluated_requirement_constraints == expected_requirement_constraints

    def test_requirement_evaluation_for_all_condition_keys(self, mocker, setup_and_teardown_injector):
        mocker.patch(
            "ahbicht.content_evaluation.rc_evaluators.RcEvaluator.evaluate_single_condition",
            side_effect=[ConditionFulfilledValue.FULFILLED, ConditionFulfilledValue.UNFULFILLED],
        )

        condition_keys = ["78", "907", "11", "583"]
        condition_node_builder = ConditionNodeBuilder(condition_keys)

        evaluated_requirement_constraints = (
            condition_node_builder.requirement_content_evaluation_for_all_condition_keys()
        )

        expected_input_nodes = {
            "11": RequirementConstraint(condition_key="11", conditions_fulfilled=ConditionFulfilledValue.UNFULFILLED),
            "78": RequirementConstraint(condition_key="78", conditions_fulfilled=ConditionFulfilledValue.FULFILLED),
            "583": self._h_583,
            "907": self._ufc_907,
        }
        assert evaluated_requirement_constraints == expected_input_nodes



"""
This module evaluates the parsed condition expression tree regarding the requirement constraints using ConditionNodes.
The RequirementConstraintTransformer defines the rules how the different parts and nodes
of the condition expression tree are handled.

The used terms are defined in the README_conditions.md.
"""

from typing import List, Literal, Mapping, Type

from lark import Token, Tree, v_args
from lark.exceptions import VisitError

from ahbicht.condition_check_results import RequirementConstraintEvaluationResult
from ahbicht.condition_node_builder import ConditionNodeBuilder, TRCTransformerArgument
from ahbicht.expressions.base_transformer import BaseTransformer
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    ConditionNode,
    EvaluatedComposition,
    FormatConstraint,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.expression_builder import FormatConstraintExpressionBuilder, HintExpressionBuilder


# pylint: disable=no-self-use
@v_args(inline=True)  # Children are provided as *args instead of a list argument
# pylint:disable=inherit-non-class
class RequirementConstraintTransformer(BaseTransformer[TRCTransformerArgument, EvaluatedComposition]):
    """
    Transformer that evaluates the trees built from the condition expressions regarding the requirement constraints.
    The input are the conditions as defined in the AHBs in the form of ConditionNodes.
    In the case of a Bedingung/RequirementConstraint already evaluated for the single condition.
    Hints are evaluated as neutral element in the boolean logic: and_composition: True, or_composition: False.

    After two nodes are evaluated in a composition, the resulting node has the node type EvalutatedComposition.
    The return value is a ConditionNode whose attribute `conditions_fulfilled` describes whether it is a required field,
    its hints are the gathered hints and its `format_constraint_expression` contains the gathered format constraints
    for the field in a newly build expression.
    """

    def and_composition(self, left: TRCTransformerArgument, right: TRCTransformerArgument) -> EvaluatedComposition:
        """Evaluates logical and_composition"""

        # if one of the nodes is neutral the condition_fulfilled.value of the other one is the resulting one
        if left.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=right.conditions_fulfilled)
        elif right.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=left.conditions_fulfilled)

        # in Python 'False and None' results in 'False' in the way we expect it,
        # but 'None and False' results in 'None', so we have to set it to False manually
        elif left.conditions_fulfilled.value is None and right.conditions_fulfilled.value is False:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=ConditionFulfilledValue(False))

        elif isinstance(left.conditions_fulfilled.value, (bool, type(None))) and isinstance(
            right.conditions_fulfilled.value, (bool, type(None))
        ):
            resulting_conditions_fulfilled = left.conditions_fulfilled.value and right.conditions_fulfilled.value
            evaluated_composition = EvaluatedComposition(
                conditions_fulfilled=ConditionFulfilledValue(resulting_conditions_fulfilled)
            )

        # Hints are added if the branch is true or neutral
        # todo: evaluated_composition might be referenced before assignment
        if evaluated_composition.conditions_fulfilled != ConditionFulfilledValue.UNFULFILLED:
            evaluated_composition.hint = (
                HintExpressionBuilder(getattr(left, "hint", None)).land(getattr(right, "hint", None)).get_expression()
            )
        evaluated_composition.format_constraints_expression = (
            FormatConstraintExpressionBuilder(left).land(right).get_expression()
        )

        return evaluated_composition

    def _or_xor_composition(
        self, left: ConditionNode, right: ConditionNode, composition: Literal["or_composition", "xor_composition"]
    ):
        """
        Determine the condition_fulfilled attribute for or_/xor_compostions.
        """
        # combining a hint with a format constraint in an or_/xor_composition has no useful result
        if (isinstance(left, Hint) and isinstance(right, UnevaluatedFormatConstraint)) or (
            isinstance(right, Hint) and isinstance(left, UnevaluatedFormatConstraint)
        ):
            raise NotImplementedError(
                f"Combining a {left.__class__} and a {right.__class__} in an"
                f"{composition} is not implemented as it has no useful result."
            )

        # combining a neutral element with a boolean value in an or_/xor_composition has no useful result
        if (
            left.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
            and isinstance(right.conditions_fulfilled.value, (bool, type(None)))
            or (
                right.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
                and isinstance(left.conditions_fulfilled.value, (bool, type(None)))
            )
        ):
            raise NotImplementedError(
                "Combining a neutral element with a boolean value in an"
                f"{composition} is not implemented as it has no useful result."
            )

        # if both nodes are neutral, the resulting one is neutral
        if (
            left.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
            and right.conditions_fulfilled == ConditionFulfilledValue.NEUTRAL
        ):
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=ConditionFulfilledValue("Neutral"))

        # in Python 'False or None' results in None the way we expect it,
        # but 'None or False' results in False, so we have to set it to None manually
        elif left.conditions_fulfilled.value is None and right.conditions_fulfilled.value is False:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=ConditionFulfilledValue(None))

        elif isinstance(left.conditions_fulfilled.value, (bool, type(None))) and isinstance(
            right.conditions_fulfilled.value, (bool, type(None))
        ):
            if composition == "or_composition":
                # todo: think of overload __or(self, ConditionFulfilledValue)__
                resulting_conditions_fulfilled = (
                    left.conditions_fulfilled.value or right.conditions_fulfilled.value
                )  # type:ignore[operator]
            elif composition == "xor_composition":
                try:
                    # todo: think of overload __xor(self, ConditionFulfilledValue)__
                    resulting_conditions_fulfilled = (
                        left.conditions_fulfilled.value ^ right.conditions_fulfilled.value  # type:ignore[operator]
                    )
                except TypeError:  # when one of the nodes is unknown because ^ does not support None
                    resulting_conditions_fulfilled = ConditionFulfilledValue(None)
            # todo: resulting_conditions_fulfilled might be referenced before assignment.
            # maybe throw not implemented exception in else branch
            evaluated_composition = EvaluatedComposition(
                conditions_fulfilled=ConditionFulfilledValue(resulting_conditions_fulfilled)
            )
        # todo: evaluated_composiiton might be referenced before assignment
        return evaluated_composition

    def or_composition(self, left: ConditionNode, right: ConditionNode) -> EvaluatedComposition:
        """Evaluates logical (inclusive) or_composition"""

        evaluated_composition = self._or_xor_composition(left, right, "or_composition")

        evaluated_composition.format_constraints_expression = (
            # todo: ask annika why the case of "invalid arguments" never happens here...
            FormatConstraintExpressionBuilder(left)  # type:ignore[arg-type]
            .lor(right)  # type:ignore[arg-type]
            .get_expression()
        )
        evaluated_composition.hint = (
            HintExpressionBuilder(getattr(left, "hint", None)).lor(getattr(right, "hint", None)).get_expression()
        )

        return evaluated_composition

    def xor_composition(self, left: ConditionNode, right: ConditionNode) -> EvaluatedComposition:
        """Evaluates exclusive xor_composition"""

        evaluated_composition = self._or_xor_composition(left, right, "xor_composition")

        evaluated_composition.format_constraints_expression = (
            # todo: ask annika why the case of "invalid arguments" never happens here...
            FormatConstraintExpressionBuilder(left)  # type:ignore[arg-type]
            .xor(right)  # type:ignore[arg-type]
            .get_expression()
        )
        evaluated_composition.hint = (
            HintExpressionBuilder(getattr(left, "hint", None)).xor(getattr(right, "hint", None)).get_expression()
        )

        return evaluated_composition

    def _then_also(
        self,
        format_constraint: UnevaluatedFormatConstraint,
        other_condition: Type[ConditionNode],
    ) -> EvaluatedComposition:
        """
        Evaluates a boolean condition with a format constraint. The functions name indicates its behaviour:
        If the `condition_fulfilled` attribute is true, the format constraint has to be fulfilled, too.
        But it is not directly evaluated but only attached to the evaluated composition such that we keep
        a strict separation of the Mussfeldprüfung itself and evaluation of format constraints on fields
        that have been marked obligatory by the (previously run) Mussfeldprüfung but do not affect
        the result of the Mussfeldprüfung.
        (Things that are obligatory are obligatory regardless of the format constraints.)
        :param format_constraint: a format constraint
        :param other_condition: a requirement constraint or hint
        :return:
        """
        if other_condition.conditions_fulfilled != ConditionFulfilledValue.NEUTRAL:
            evaluated_composition = EvaluatedComposition(conditions_fulfilled=other_condition.conditions_fulfilled)
            format_constraint_is_required = other_condition.conditions_fulfilled.value
        elif isinstance(other_condition, Hint):
            evaluated_composition = EvaluatedComposition(
                conditions_fulfilled=ConditionFulfilledValue("Neutral"), hint=other_condition.hint
            )
            format_constraint_is_required = True
        else:
            raise NotImplementedError(
                f"Combining Conditions of type {other_condition.__class__} with a "
                f"{format_constraint.__class__} is not implemented."
            )
        if format_constraint_is_required:
            evaluated_composition.format_constraints_expression = (
                FormatConstraintExpressionBuilder(format_constraint).land(other_condition).get_expression()
            )

        return evaluated_composition

    def then_also_composition(self, left: Type[ConditionNode], right: Type[ConditionNode]) -> EvaluatedComposition:
        """
        A "then also" composition is typically used for format constraints.
        It connects an evaluable expression with a format constraint.
        If the `conditions_fulfilled` attribute evaluates to true,
        then also the attached format constraint has to be fulfilled.
        """
        # pylint: disable=protected-access
        if isinstance(left, UnevaluatedFormatConstraint):
            return self._then_also(format_constraint=left, other_condition=right)
        return self._then_also(
            format_constraint=right,  # type:ignore[arg-type]
            # #because typically format constraints are "attached" to the right side of an expression
            other_condition=left,
        )  # this might raise the NotImplementedError


def evaluate_requirement_constraint_tree(
    parsed_tree: Tree, input_values: Mapping[str, TRCTransformerArgument]
) -> EvaluatedComposition:
    """
    Evaluates the tree built from the expressions with the help of the ConditionsTransformer.

    :param parsed_tree: Tree
    :param input_values: dict(condition_key, ConditionNode)
        :return: EvaluatedComposition
    """
    if not all(isinstance(input_value, ConditionNode) for input_value in input_values.values()):
        raise ValueError("Please make sure that the passed values are ConditionNodes.")
    if not all(
        isinstance(input_value, (RequirementConstraint, Hint, FormatConstraint))
        for input_value in input_values.values()
    ):
        raise ValueError(
            """Please make sure that the passed values are ConditionNodes \
of the type RequirementConstraint, Hint or FormatConstraint."""
        )
    try:
        result = RequirementConstraintTransformer(input_values).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc

    return result


def requirement_constraint_evaluation(condition_expression: str) -> RequirementConstraintEvaluationResult:
    """
    Evaluation of the condition expression in regard to the requirement conditions (rc).
    """

    parsed_tree_rc: Tree = parse_condition_expression_to_tree(condition_expression)

    # get all condition keys from tree
    all_condition_keys: List[str] = [
        t.value for t in parsed_tree_rc.scan_values(lambda v: isinstance(v, Token))  # type: ignore[attr-defined]
    ]
    condition_node_builder = ConditionNodeBuilder(all_condition_keys)
    input_nodes = condition_node_builder.requirement_content_evaluation_for_all_condition_keys()

    resulting_condition_node: ConditionNode = evaluate_requirement_constraint_tree(parsed_tree_rc, input_nodes)

    requirement_constraints_fulfilled = resulting_condition_node.conditions_fulfilled.value
    requirement_is_conditional = True
    if resulting_condition_node.conditions_fulfilled == ConditionFulfilledValue("Neutral"):  # pylint:disable=no-member
        requirement_constraints_fulfilled = True
        requirement_is_conditional = False
    if resulting_condition_node.conditions_fulfilled == ConditionFulfilledValue(None):  # pylint:disable=no-member
        raise NotImplementedError("It is unknown if the conditions are fulfilled due to missing information.")

    format_constraints_expression = getattr(resulting_condition_node, "format_constraints_expression", None)
    if isinstance(resulting_condition_node, UnevaluatedFormatConstraint):
        format_constraints_expression = f"[{resulting_condition_node.condition_key}]"
    hints = getattr(resulting_condition_node, "hint", None)

    requirement_constraint_evaluation_result = RequirementConstraintEvaluationResult(
        requirement_constraints_fulfilled=requirement_constraints_fulfilled,
        hints=hints,
        format_constraints_expression=format_constraints_expression,
        requirement_is_conditional=requirement_is_conditional,
    )

    return requirement_constraint_evaluation_result



"""
Module for taking all the condition keys of a condition expression and building their respective ConditionNodes.
If necessary it evaluates the needed attributes.
"""
import asyncio
from typing import Dict, List, Tuple, Union

import inject

from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.expressions.condition_nodes import Hint, RequirementConstraint, UnevaluatedFormatConstraint
from ahbicht.expressions.hints_provider import HintsProvider

# TRCTransformerArgument is a union of nodes that are already evaluated from a Requirement Constraint (RC) perspective.
# The Format Constraints (FC) might still be unevaluated. That's why the return type used in the
# RequirementConstraintTransformer is always an EvaluatedComposition.
TRCTransformerArgument = Union[RequirementConstraint, UnevaluatedFormatConstraint, Hint]


# pylint: disable=no-member, too-few-public-methods


class ConditionNodeBuilder:
    """
    Builds ConditionNodes for the given condition_keys by separating them into their respective types
    and evaluating the necessary attributes.
    It distinguishes between requirement constraint evaluation and format constraint evaluation.
    """

    def __init__(self, condition_keys: List[str]):
        self.hints_provider: HintsProvider = inject.instance(HintsProvider)  # type:ignore[assignment]
        self.rc_evaluator: RcEvaluator = inject.instance(RcEvaluator)  # type:ignore[assignment]
        self.condition_keys = condition_keys
        (
            self.requirement_constraints_condition_keys,
            self.hints_condition_keys,
            self.format_constraints_condition_keys,
        ) = self._seperate_condition_keys_into_each_type()

    def _seperate_condition_keys_into_each_type(self) -> Tuple[List[str], List[str], List[str]]:
        """
        Separates the list of all condition keys into three lists of their respective types.
        The types are differentiated by their number range.
        See 'Allgemeine Festlegungen' from EDI@Energy.
        """
        requirement_constraints: List[str] = []
        hints: List[str] = []
        format_constraints: List[str] = []
        for condition_key in self.condition_keys:
            if 1 <= int(condition_key) <= 499:
                requirement_constraints.append(condition_key)
            elif 500 <= int(condition_key) <= 900:
                hints.append(condition_key)
            elif 901 <= int(condition_key) <= 999:
                format_constraints.append(condition_key)
            else:
                raise ValueError("Condition key is not in valid number range.")
        return requirement_constraints, hints, format_constraints

    def _build_hint_nodes(self) -> Dict[str, Hint]:
        """Builds Hint nodes from their condition keys by getting all hint texts from the HintsProvider."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(self.hints_provider.get_hints(self.hints_condition_keys))

    def _build_unevaluated_format_constraint_nodes(self) -> Dict[str, UnevaluatedFormatConstraint]:
        """Build unevaluated format constraint nodes."""
        unevaluated_format_constraints: Dict[str, UnevaluatedFormatConstraint] = {}
        for condition_key in self.format_constraints_condition_keys:
            unevaluated_format_constraints[condition_key] = UnevaluatedFormatConstraint(condition_key=condition_key)
        return unevaluated_format_constraints

    def _build_requirement_constraint_nodes(self) -> Dict[str, RequirementConstraint]:
        """
        Build requirement constraint nodes by evaluating the constraints
        with the help of the respective Evaluator.
        """

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        evaluated_conditions_fulfilled_attribute: dict = loop.run_until_complete(
            self.rc_evaluator.evaluate_conditions(self.requirement_constraints_condition_keys)
        )

        evaluated_requirement_constraints: Dict[str, RequirementConstraint] = {}
        for condition_key in self.requirement_constraints_condition_keys:
            evaluated_requirement_constraints[condition_key] = RequirementConstraint(
                condition_key=condition_key,
                conditions_fulfilled=evaluated_conditions_fulfilled_attribute[condition_key],
            )
        return evaluated_requirement_constraints

    def requirement_content_evaluation_for_all_condition_keys(
        self,
    ) -> Dict[str, TRCTransformerArgument]:
        """Gets input nodes for all condition keys."""
        requirement_constraint_nodes = self._build_requirement_constraint_nodes()
        hint_nodes = self._build_hint_nodes()
        unevaluated_format_constraint_nodes = self._build_unevaluated_format_constraint_nodes()
        input_nodes: Dict[str, TRCTransformerArgument] = {
            **requirement_constraint_nodes,
            **hint_nodes,
            **unevaluated_format_constraint_nodes,
        }
        return input_nodes



""" Tests for the parsing of the conditions tests (Mussfeldprüfung) """

import pytest
from lark import Token, Tree

from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree


class TestConditionParser:
    """Test for the parsing of the conditions tests (Mussfeldprüfung)"""

    @pytest.mark.parametrize(
        "expression, expected_tree",
        [
            pytest.param(
                # single condition
                "[1]",
                Tree("condition_key", [Token("INT", "1")]),
            ),
            pytest.param(
                # single condition with whitespace
                "[1  ]",
                Tree("condition_key", [Token("INT", "1")]),
            ),
            pytest.param(
                # simple or_composition
                "[1]O[2]",
                Tree(
                    "or_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
            pytest.param(
                # simple or_composition with whitespace
                " [1] O[ 2]",
                Tree(
                    "or_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
            pytest.param(
                # simple or_composition with whitespace and tab
                " [1]\tO[ 2]",
                Tree(
                    "or_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
            pytest.param(
                # and/or combination, and before or
                "[1]U[2]    O[53]",
                Tree(
                    "or_composition",
                    [
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree("condition_key", [Token("INT", "53")]),
                    ],
                ),
            ),
            pytest.param(
                # and/or combination, and before or, different order
                "[53]O[1]U[2]",
                Tree(
                    "or_composition",
                    [
                        Tree("condition_key", [Token("INT", "53")]),
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # xor_composition
                "[1]X[2]",
                Tree(
                    "xor_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
            pytest.param(
                # xor_composition
                "[1]⊻[2]",
                Tree(
                    "xor_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
        ],
    )
    def test_parse_valid_expression_to_tree(self, expression: str, expected_tree: Tree):
        """
        Tests that valid expressions containing operators "O"/"U"/"X", different whitespaces
        and no brackets are parsed as expected.
        """
        parsed_tree = parse_condition_expression_to_tree(expression)

        assert isinstance(parsed_tree, Tree)
        assert parsed_tree == expected_tree

    @pytest.mark.parametrize(
        "expression, expected_tree",
        [
            pytest.param(
                #  no brackets
                " [1][987]",
                Tree(
                    "then_also_composition",
                    [
                        Tree("condition_key", [Token("INT", "1")]),
                        Tree("condition_key", [Token("INT", "987")]),
                    ],
                ),
            ),
            pytest.param(
                #  two format constraints with an operator
                "[901]U[987]",
                Tree(
                    "and_composition",
                    [
                        Tree("condition_key", [Token("INT", "901")]),
                        Tree("condition_key", [Token("INT", "987")]),
                    ],
                ),
            ),
            pytest.param(
                # format constraint is attached as suffix _without_ an operator
                "([1]U[2])[987]",
                Tree(
                    "then_also_composition",
                    [
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree("condition_key", [Token("INT", "987")]),
                    ],
                ),
            ),
            pytest.param(
                # format constraint is attached as prefix _without_ an operator
                "[987]([1]U[2])",
                Tree(
                    "then_also_composition",
                    [
                        Tree("condition_key", [Token("INT", "987")]),
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # MSCONS AHB, Kapitel 7
                "([902] U [906] [46])",
                Tree(
                    "and_composition",
                    [
                        Tree("condition_key", [Token("INT", "902")]),
                        Tree(
                            "then_also_composition",
                            [
                                Tree("condition_key", [Token("INT", "906")]),
                                Tree("condition_key", [Token("INT", "46")]),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # two format constraints
                "([950]([2]U[4]))O([951]([1]U[3]))",
                Tree(
                    "or_composition",
                    [
                        Tree(
                            "then_also_composition",
                            [
                                Tree("condition_key", [Token("INT", "950")]),
                                Tree(
                                    "and_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "2")]),
                                        Tree("condition_key", [Token("INT", "4")]),
                                    ],
                                ),
                            ],
                        ),
                        Tree(
                            "then_also_composition",
                            [
                                Tree("condition_key", [Token("INT", "951")]),
                                Tree(
                                    "and_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "1")]),
                                        Tree("condition_key", [Token("INT", "3")]),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ),
        ],
    )
    def test_parse_valid_expression_to_tree_with_format_constraints(self, expression: str, expected_tree: Tree):
        """
        Tests that valid expressions containing operators "O" and "U", different whitespaces
        and no brackets are parsed as expected. It is similiar to the test `test_parse_valid_expression_to_tree`
        but it adds format constraints.
        """
        parsed_tree = parse_condition_expression_to_tree(expression)

        assert isinstance(parsed_tree, Tree)
        assert parsed_tree == expected_tree

    @pytest.mark.parametrize(
        "expression, expected_tree",
        [
            pytest.param(
                # single condition
                "([1])",
                Tree("condition_key", [Token("INT", "1")]),
            ),
            pytest.param(
                # simple or_composition
                "([1]O[2])",
                Tree(
                    "or_composition",
                    [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                ),
            ),
            pytest.param(
                # and/or combination, and in brackets
                "([1]U[2])O[53]",
                Tree(
                    "or_composition",
                    [
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree("condition_key", [Token("INT", "53")]),
                    ],
                ),
            ),
            pytest.param(
                # and/or combination, or in brackets
                "([1]O[2])U[53]",
                Tree(
                    "and_composition",
                    [
                        Tree(
                            "or_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree("condition_key", [Token("INT", "53")]),
                    ],
                ),
            ),
            pytest.param(
                # and/or combination, or in brackets, different order
                "[53]U([1]O[2])",
                Tree(
                    "and_composition",
                    [
                        Tree("condition_key", [Token("INT", "53")]),
                        Tree(
                            "or_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # complex expression with two brackets
                "([1]O[2])U([53]U[4]O[12])",
                Tree(
                    "and_composition",
                    [
                        Tree(
                            "or_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree(
                            "or_composition",
                            [
                                Tree(
                                    "and_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "53")]),
                                        Tree("condition_key", [Token("INT", "4")]),
                                    ],
                                ),
                                Tree("condition_key", [Token("INT", "12")]),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # complex expression with two brackets
                "([1]∨[2])∧([53]∧[4]∨[12])",
                Tree(
                    "and_composition",
                    [
                        Tree(
                            "or_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                        Tree(
                            "or_composition",
                            [
                                Tree(
                                    "and_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "53")]),
                                        Tree("condition_key", [Token("INT", "4")]),
                                    ],
                                ),
                                Tree("condition_key", [Token("INT", "12")]),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                # nested brackets
                "[100]U([2]U([53]O[4]))",
                Tree(
                    "and_composition",
                    [
                        Tree("condition_key", [Token("INT", "100")]),
                        Tree(
                            "and_composition",
                            [
                                Tree("condition_key", [Token("INT", "2")]),
                                Tree(
                                    "or_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "53")]),
                                        Tree("condition_key", [Token("INT", "4")]),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ),
        ],
    )
    def test_parse_valid_expression_with_brackets_to_tree(self, expression: str, expected_tree: Tree):
        """Tests that valid strings that contain brackets are parsed as expected."""
        parsed_tree = parse_condition_expression_to_tree(expression)

        assert isinstance(parsed_tree, Tree)
        assert parsed_tree == expected_tree

    @pytest.mark.parametrize(
        "expression",
        [
            pytest.param(None),
            pytest.param(""),
            pytest.param("   "),
            pytest.param("[1"),
            pytest.param("1]"),
            pytest.param("1"),
            pytest.param("[]"),
            pytest.param("[2]U[1"),
            pytest.param("[2]U1"),
            pytest.param("[2]U"),
            pytest.param("([1]U[2]"),  # missing closing bracket
            pytest.param("[1]U[2])"),  # missing opening bracket
        ],
    )
    def test_parse_invalid_expression(self, expression: str):
        """Tests that an error is raised when trying to parse an invalid string."""

        with pytest.raises(SyntaxError) as excinfo:
            parse_condition_expression_to_tree(expression)

        assert """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """ in str(
            excinfo.value
        )

    @pytest.mark.parametrize(
        "old_expression, new_expression",
        [
            pytest.param("[1]U[2]", "[1]∧[2]"),
            pytest.param("[1]O[2]", "[1]∨[2]"),
            pytest.param("[1]X[2]", "[1]⊻[2]"),
            pytest.param(
                "([951][510]U[522])O([950][514]U([523]O[525]))", "([951][510]∧[522])∨([950][514]∧([523]∨[525]))"
            ),
        ],
    )
    def test_equivalence_of_new_and_old_notation_expressions(self, old_expression: str, new_expression: str):
        """
        Tests that U/O/X are treated just the same as the new logical operands.
        :return:
        """
        old_tree = parse_condition_expression_to_tree(old_expression)
        new_tree = parse_condition_expression_to_tree(new_expression)
        assert old_tree == new_tree



""" Test for the requirement constraint evaluation of the condition expressions. """

from unittest.mock import AsyncMock

import inject
import pytest

from ahbicht.condition_check_results import RequirementConstraintEvaluationResult
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedComposition,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.hints_provider import HintsProvider
from ahbicht.expressions.requirement_constraint_expression_evaluation import requirement_constraint_evaluation


class TestRequirementConstraintEvaluation:
    """Test for the evaluation of the condition expression regarding the requirement constraints."""

    _input_values = {
        "1": RequirementConstraint(condition_key="1", conditions_fulfilled=ConditionFulfilledValue.FULFILLED),
        "2": RequirementConstraint(condition_key="2", conditions_fulfilled=ConditionFulfilledValue.UNFULFILLED),
        "3": RequirementConstraint(condition_key="3", conditions_fulfilled=ConditionFulfilledValue.FULFILLED),
        "4": RequirementConstraint(condition_key="4", conditions_fulfilled=ConditionFulfilledValue.UNFULFILLED),
        "101": RequirementConstraint(condition_key="101", conditions_fulfilled=ConditionFulfilledValue.UNKNOWN),
        "901": UnevaluatedFormatConstraint(condition_key="901"),
        "902": UnevaluatedFormatConstraint(condition_key="902"),
        "503": Hint(condition_key="503", hint="[503] Hinweis:foo"),
        "504": Hint(condition_key="504", hint="[504] Hinweis:bar"),
    }

    @pytest.fixture()
    def setup_and_teardown_injector(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(HintsProvider, AsyncMock(wraps=HintsProvider)).bind(
                RcEvaluator, AsyncMock(wraps=RcEvaluator)
            )
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        """condition_expression, expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional, expected_format_constraints_expression, expected_hints""",
        [
            pytest.param("[1]", True, True, None, None),
            pytest.param("[2]", False, True, None, None),
            pytest.param("[2]\t", False, True, None, None),
            pytest.param(" [ 1]  ", True, True, None, None),
            pytest.param("[4\t]", False, True, None, None),
            pytest.param("[1]U[2]", False, True, None, None),
            pytest.param("[1]U[2]U[3]O[4]", False, True, None, None),
            pytest.param("([1]O[2])U[3]", True, True, None, None),
            pytest.param("[2]O[4]", False, True, None, None),
            # Neutral value
            pytest.param("[503]", True, False, None, "[503] Hinweis:foo"),
            pytest.param("[902]", True, False, "[902]", None),
            pytest.param("[503]U[504]", True, False, None, "[503] Hinweis:foo und [504] Hinweis:bar"),
        ],
    )
    def test_evaluate_valid_ahb_expression(
        self,
        mocker,
        condition_expression,
        expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional,
        expected_format_constraints_expression,
        expected_hints,
        setup_and_teardown_injector,
    ):
        """
        Tests that valid ahb expressions are evaluated as expected.
        Odd condition_keys are True, even condition_keys are False
        """
        mocker.patch(
            "ahbicht.expressions.requirement_constraint_expression_evaluation.ConditionNodeBuilder.requirement_content_evaluation_for_all_condition_keys",
            return_value=self._input_values,
        )
        requirement_constraint_evaluation_result = requirement_constraint_evaluation(
            condition_expression=condition_expression
        )

        assert isinstance(requirement_constraint_evaluation_result, RequirementConstraintEvaluationResult)

        assert (
            requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            == expected_requirement_constraints_fulfilled
        )
        assert (
            requirement_constraint_evaluation_result.requirement_is_conditional == expected_requirement_is_conditional
        )
        assert (
            requirement_constraint_evaluation_result.format_constraints_expression
            == expected_format_constraints_expression
        )
        assert requirement_constraint_evaluation_result.hints == expected_hints

    @pytest.mark.parametrize(
        "condition_expression, input_values, expected_error, expected_error_message",
        [
            pytest.param(
                "[]",
                {"1": "no_boolean"},
                SyntaxError,
                """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """,
            ),
            pytest.param(
                "[1]U[2]",
                {"1": "no_boolean", "2": ConditionFulfilledValue.FULFILLED},
                ValueError,
                "Please make sure that the passed values are ConditionNodes.",
            ),
            pytest.param(
                "[1]",
                {"1": EvaluatedComposition(conditions_fulfilled=ConditionFulfilledValue.FULFILLED)},
                ValueError,
                "Please make sure that the passed values are ConditionNodes of the type RequirementConstraint, "
                "Hint or FormatConstraint.",
            ),
            # no value for [2]
            pytest.param(
                "[1]U[2]",
                {"1": RequirementConstraint(condition_key="1", conditions_fulfilled=ConditionFulfilledValue.FULFILLED)},
                ValueError,
                "Please make sure that the input values contain all necessary condition_keys.",
            ),
            # unknown value as result
            pytest.param(
                "[101]",
                _input_values,
                NotImplementedError,
                "It is unknown if the conditions are fulfilled due to missing information.",
            ),
        ],
    )
    def test_evaluate_condition_expression_with_invalid_values(
        self,
        mocker,
        condition_expression: str,
        input_values: dict,
        expected_error: type,
        expected_error_message: str,
        setup_and_teardown_injector,
    ):
        """Tests that an error is raised when trying to pass invalid values."""
        mocker.patch(
            "ahbicht.expressions.requirement_constraint_expression_evaluation.ConditionNodeBuilder.requirement_content_evaluation_for_all_condition_keys",
            return_value=input_values,
        )
        with pytest.raises(expected_error) as excinfo:
            requirement_constraint_evaluation(condition_expression)

        assert expected_error_message in str(excinfo.value)



"""
Tests the hints provider module.
"""
import pytest

from ahbicht.edifact import EdifactFormat, EdifactFormatVersion
from ahbicht.expressions.hints_provider import JsonFileHintsProvider

pytestmark = pytest.mark.asyncio


class TestHintsProvider:
    """Test Class for JsonFileHintsProvider"""

    @pytest.mark.datafiles("./unittests/resources_condition_hints/FV2104/Hints_FV2104_UTILMD.json")
    async def test_initiating_hints_provider(self, datafiles):
        """Tests if hints provider is initiated correctly."""
        path_to_hint_json = datafiles / "Hints_FV2104_UTILMD.json"
        hints_provider = JsonFileHintsProvider(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            file_path=path_to_hint_json,
        )
        assert hints_provider.edifact_format == EdifactFormat.UTILMD
        assert hints_provider.edifact_format_version == EdifactFormatVersion.FV2104
        assert await hints_provider.get_hint_text("583") == "[583] Hinweis: Verwendung der ID der Marktlokation"



"""
A format constraint (FC) evaluator is an evaluator for format constraints.
Think of stuff like:
* strings that should match an OBIS regex
* MarktlokationsIDs having correct check sums
* pre/post decimal values having specific ranges
Other than requirement constraints format constraints do not affect if data are required at all, but instead only
validate already required data.
"""
import asyncio
from abc import ABC
from typing import Coroutine, Dict, List

from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint


# pylint: disable=no-self-use
class FcEvaluator(Evaluator, ABC):
    """
    Base of all Format Constraint (FC) evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    async def evaluate_single_format_constraint(
        self, condition_key: str, entered_input: str
    ) -> EvaluatedFormatConstraint:
        """
        Evaluates the format constraint with the given key.
        :param condition_key: key of the condition, e.g. "950"
        :param entered_input: the entered input whose format should be checked, e.g. "12345678913"
        :return: If the format constraint is fulfilled and an optional error_message.
        """
        evaluation_method = self.get_evaluation_method(condition_key)
        if evaluation_method is None:
            raise NotImplementedError(f"There is no content_evaluation method for format constraint '{condition_key}'")
        result: EvaluatedFormatConstraint = await evaluation_method(entered_input)

        # Fallback error message if there is no error message even though format constraint isn't fulfilled
        if result.format_constraint_fulfilled is False and result.error_message is None:
            result.error_message = f"Condition [{condition_key}] has to be fulfilled."

        return result

    async def evaluate_format_constraints(
        self, condition_keys: List[str], entered_input: str
    ) -> Dict[str, EvaluatedFormatConstraint]:
        """
        Evaluate the entered_input in regard to all the formats provided in condition_keys.
        """
        tasks: List[Coroutine] = [
            self.evaluate_single_format_constraint(condition_key, entered_input) for condition_key in condition_keys
        ]
        results: List[EvaluatedFormatConstraint] = await asyncio.gather(*tasks)

        result: Dict[str, EvaluatedFormatConstraint] = dict(zip(condition_keys, results))
        return result


class DictBasedFcEvaluator(FcEvaluator):
    """
    A format constraint evaluator that is initialized with a prefilled dictionary.
    """

    def __init__(self, results: Dict[str, EvaluatedFormatConstraint]):
        """
        Initialize with a dictionary that contains all the format constraint evaluation results.
        :param results:
        """
        self._results: Dict[str, EvaluatedFormatConstraint] = results

    # pylint: disable=unused-argument
    async def evaluate_single_format_constraint(
        self, condition_key: str, entered_input: str
    ) -> EvaluatedFormatConstraint:
        try:
            return self._results[condition_key]
        except KeyError as key_error:
            raise NotImplementedError(f"No result was provided for {condition_key}.") from key_error



"""
This module parses a condition expression like "[59] U ([123] O [456])" into a tree structur using
the parsing library lark: https://lark-parser.readthedocs.io/en/latest/

The used terms are defined in the README_conditions.md.
"""

from lark import Lark, Tree
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF


def parse_condition_expression_to_tree(condition_expression: str) -> Tree:
    """
    Parse a given condition expression with the help of the here defined grammar to a lark tree.
    The grammar starts with condition keys, e.g. [45] and combines them with
    and _/or_compositions corresponding to U/O operators or without an operator (then_also_composition).
    It follows the boolean logic 'brackets before `then_also` before `and` before `xor` before `or`'.
    Whitespaces are ignored.

    :param condition_expression: str, e.g. '[45]U[502]O[1][906]'
    :return parsed_tree: Tree
    """

    grammar = """
    ?expression: expression "O" expression -> or_composition
                | expression "∨" expression -> or_composition
                | expression "X" expression -> xor_composition
                | expression "⊻" expression -> xor_composition
                | expression "U" expression -> and_composition
                | expression "∧" expression -> and_composition
                | expression expression -> then_also_composition
                | brackets
                | condition_key
    ?brackets: "(" expression ")"
    condition_key: "[" INT "]"

    %import common.INT
    %import common.WS
    %ignore WS  // WS = whitespace
    """

    parser = Lark(grammar, start="expression")
    try:
        parsed_tree = parser.parse(condition_expression)
    except (UnexpectedEOF, UnexpectedCharacters, TypeError) as eof:
        raise SyntaxError(
            """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """
        ) from eof

    return parsed_tree



"""
This module manages EDIFACT related stuff. It's basically a helper module to avoid stringly typed parameters.
"""
import re
from typing import Optional

import aenum  # type: ignore[import]

pruefidentifikator_pattern = re.compile(r"^[1-9]\d{4}$")


# pylint: disable=too-few-public-methods
class EdifactFormat(aenum.Enum):
    """
    existing EDIFACT formats
    """

    _init_ = "value string"
    APERAK = 99, "APERAK"
    IFTSTA = 21, "IFTSTA"  # Multimodaler Statusbericht
    INSRPT = 23, "INSRPT"  # Prüfbericht
    INVOIC = 31, "INVOIC"  # invoice
    MSCONS = 13, "MSCONS"  # meter readings
    ORDERS = 17, "ORDERS"  # orders
    ORDRSP = 19, "ORDRSP"  # orders response
    PRICAT = 21, "PRICAT"  # price catalogue
    QUOTES = 15, "QUOTES"  # quotes
    REMADV = 33, "REMADV"  # zahlungsavis
    REQOTE = 35, "REQOTE"  # request quote
    UTILMD = 11, "UTILMD"  # utilities master data
    UTILTS = 25, "UTILTS"  # formula

    def __str__(self):
        return self.string


class EdifactFormatVersion(aenum.Enum):
    """
    One format version refers to the period in which an AHB is valid.
    """

    _init_ = "value string"
    FV2104 = 2104, "FV2104"  # valid since 2021-04-01
    FV2110 = 2110, "FV2110"  # valid from 2021-10-01 onwards
    FV2204 = 2204, "FV2204"  # valid from 2022-04-01 onwards ("MaKo 2022")

    def __str__(self):
        return self.string


def pruefidentifikator_to_format(pruefidentifikator: str) -> Optional[EdifactFormat]:
    """
    returns the format corresponding to a given pruefi
    :param pruefidentifikator:
    :return: matching EDIFACT format or None
    """
    if not pruefidentifikator:
        raise ValueError("The pruefidentifikator must not be falsy")
    if not pruefidentifikator_pattern.match(pruefidentifikator):
        raise ValueError(f"The pruefidentifikator '{pruefidentifikator}' is invalid.")
    try:
        result: EdifactFormat = EdifactFormat(int(pruefidentifikator[:2]))
        return result
    except ValueError:
        return None



"""
Module to provide the condition hints from their respective json file
as dictionary with the condition keys as keys and the hint texts as values.
"""
import asyncio
import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Mapping, Optional

from ahbicht.edifact import EdifactFormat, EdifactFormatVersion

# pylint: disable = too-few-public-methods
from ahbicht.expressions.condition_nodes import Hint


class HintsProvider(ABC):
    """
    A Hints Provider provides plain text hints for a given condition number.
    This class provides the hints from the respective json file (defined by EdifactFormatVersion and EdifactFormat)
    as dictionary with the condition keys as keys and the hint texts as values.
    """

    edifact_format: EdifactFormat = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format to which it is applicable."
    )

    edifact_format_version: EdifactFormatVersion = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format version."
    )

    @abstractmethod
    async def get_hint_text(self, condition_key: str) -> Optional[str]:
        """
        Get the hint text for the given condition key.
        :param condition_key: e.g. "501"
        :return: the corresponding hint text, e.g. "Data is thought to be interpreted as foo bar."
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    async def get_hints(self, condition_keys: List[str], raise_key_error: bool = True) -> Dict[str, Hint]:
        """
        Get Hints for given condition keys by asynchronously awaiting all self.get_hint_text at once
        """
        tasks = [self.get_hint_text(ck) for ck in condition_keys]
        results: List[Optional[str]] = await asyncio.gather(*tasks)
        result: Dict[str, Hint] = {}
        for key, value in zip(condition_keys, results):
            if value is None:
                if raise_key_error:
                    raise KeyError(f"There seems to be no hint implemented with condition key '{key}'.")
            else:
                result[key] = Hint(hint=value, condition_key=key)
        return result


class DictBasedHintsProvider(HintsProvider):
    """
    A Hints Provider that is based on hardcoded values from a dictionary
    """

    def __init__(self, results: Mapping[str, Optional[str]]):
        """
        Initialize with a dictionary that contains all the Hinweis texts.
        :param results:
        """
        self._all_hints: Mapping[str, Optional[str]] = results

    async def get_hint_text(self, condition_key: str) -> Optional[str]:
        if not condition_key:
            raise ValueError(f"The condition key must not be None/empty but was '{condition_key}'")
        if condition_key in self._all_hints:
            return self._all_hints[condition_key]
        return None


class JsonFileHintsProvider(DictBasedHintsProvider):
    """
    The JsonFileHintsProvider loads hints from a JSON file.
    """

    def __init__(self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, file_path: Path):
        super().__init__(self._open_and_load_hint_json(file_path))
        self.edifact_format = edifact_format
        self.edifact_format_version = edifact_format_version

    @staticmethod
    def _open_and_load_hint_json(file_path: Path) -> Dict[str, str]:
        """
        Opens the hint json file and loads it into an attribute of the class.
        """
        with open(file_path, encoding="utf-8") as json_infile:
            return json.load(json_infile)



""" Tests for creating different condition nodes that are used in the parsed tree. """

import pytest

from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    ConditionNode,
    EvaluatedComposition,
    Hint,
    RequirementConstraint,
)


class TestConditionNodes:
    """Test for creating different condition nodes that are used in the parsed tree."""

    @pytest.mark.parametrize(
        "cfv, equivalent_string",
        [
            pytest.param(
                ConditionFulfilledValue.FULFILLED,
                "FULFILLED",
            ),
        ],
    )
    def test_condition_fulfilled_value_equality(self, cfv: ConditionFulfilledValue, equivalent_string: str):
        """For mypy we had to replace some enum comparisons. This test is to ensure that everything works as expected"""
        assert cfv == ConditionFulfilledValue(ConditionFulfilledValue.FULFILLED)

    @pytest.mark.parametrize(
        "conditions_fulfilled, conditions_fulfilled_expected_value",
        [
            pytest.param(ConditionFulfilledValue.FULFILLED, True),
            pytest.param(ConditionFulfilledValue.UNFULFILLED, False),
            pytest.param(ConditionFulfilledValue.NEUTRAL, "Neutral"),
        ],
    )
    def test_valid_requirement_constraint(self, conditions_fulfilled, conditions_fulfilled_expected_value):
        """Tests the creation of a valid requirement constraint node."""
        actual_node = RequirementConstraint(
            condition_key="1",
            conditions_fulfilled=conditions_fulfilled,
        )

        assert isinstance(actual_node, ConditionNode)
        assert isinstance(actual_node, RequirementConstraint)
        assert actual_node.condition_key == "1"
        assert actual_node.conditions_fulfilled is conditions_fulfilled
        assert actual_node.conditions_fulfilled.value is conditions_fulfilled_expected_value

    @pytest.mark.parametrize(
        "condition_node_arguments, expected_error_message",
        [
            pytest.param(
                {"conditions_fulfilled": ConditionFulfilledValue.FULFILLED},
                "missing 1 required keyword-only argument",
            ),
            pytest.param(
                {
                    "condition_key": "1",
                    "conditions_fulfilled": ConditionFulfilledValue.FULFILLED,
                    "hint": "foo",
                },
                "got an unexpected keyword argument",
            ),
            pytest.param(
                {
                    "condition_key": "1",
                    "conditions_fulfilled": "no_ConditionFulfilledValue",
                },
                f"'conditions_fulfilled' must be <aenum '{ConditionFulfilledValue.__name__}'>",
            ),
        ],
    )
    def test_invalid_requirement_constraint(self, condition_node_arguments, expected_error_message):
        """Tests if requirements for RequirementConstraints are working as expected."""
        with pytest.raises(TypeError) as excinfo:
            RequirementConstraint(**condition_node_arguments)
        assert expected_error_message in str(excinfo.value)

    def test_valid_hint(self):
        """Tests the creation of a valid Hint node."""
        actual_node = Hint(condition_key="501", hint="[501] Hinweis: Foo")

        assert isinstance(actual_node, ConditionNode)
        assert isinstance(actual_node, Hint)
        assert actual_node.condition_key == "501"
        assert actual_node.hint == "[501] Hinweis: Foo"

    @pytest.mark.parametrize(
        "hint_node_arguments, expected_error_message",
        [
            pytest.param(
                {"hint": "[501] Hinweis: Foo"},
                "missing 1 required keyword-only argument",
            ),
            pytest.param(
                {"condition_key": "1"},
                "missing 1 required keyword-only argument",
            ),
            pytest.param(
                {
                    "condition_key": "501",
                    "hint": "[501] Hinweis: Foo",
                    "format": "bar",
                },
                "got an unexpected keyword argument",
            ),
            pytest.param(
                {"condition_key": "501", "hint": None},
                "'hint' must be <class 'str'>",
            ),
        ],
    )
    def test_invalid_hint(self, hint_node_arguments, expected_error_message):
        """Tests if requirements for Hints are working as expected."""

        with pytest.raises(TypeError) as excinfo:
            Hint(**hint_node_arguments)
        assert expected_error_message in str(excinfo.value)

    def test_valid_evaluated_composition(self):
        """Tests the creation of a valid EvaluatedComposition."""
        # Minimal example:
        minimal_node = EvaluatedComposition(conditions_fulfilled=ConditionFulfilledValue.FULFILLED)

        assert isinstance(minimal_node, ConditionNode)
        assert isinstance(minimal_node, EvaluatedComposition)
        assert minimal_node.conditions_fulfilled.value is True

        maximal_node = EvaluatedComposition(
            conditions_fulfilled=ConditionFulfilledValue.UNFULFILLED,
            hint="[501] Hinweis: Foo",
        )

        assert isinstance(maximal_node, ConditionNode)
        assert isinstance(maximal_node, EvaluatedComposition)
        assert maximal_node.conditions_fulfilled.value is False
        assert maximal_node.hint == "[501] Hinweis: Foo"

    @pytest.mark.parametrize(
        "resulting_node_arguments, expected_error_message",
        [
            pytest.param(
                {},
                "missing 1 required keyword-only argument",
            ),
            pytest.param(
                {
                    "condition_key": "501",
                    "conditions_fulfilled": True,
                },
                "got an unexpected keyword argument",
            ),
            pytest.param(
                {
                    "conditions_fulfilled": "no_ConditionFulfilledValue",
                },
                f"'conditions_fulfilled' must be <aenum '{ConditionFulfilledValue.__name__}'>",
            ),
        ],
    )
    def test_invalid_evaluated_composition(self, resulting_node_arguments, expected_error_message):
        """Tests if requirements for Hints are working as expected."""
        with pytest.raises(TypeError) as excinfo:
            EvaluatedComposition(**resulting_node_arguments)
        assert expected_error_message in str(excinfo.value)



"""
This module evaluates the parsed ahb expression tree.
The AhbExpressionTransformer defines the rules how the different parts of the parsed tree are handled.

The used terms are defined in the README.md.
"""

from typing import List

from lark import Token, Transformer, Tree, v_args
from lark.exceptions import VisitError

from ahbicht.condition_check_results import (
    ConditionCheckResult,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions.format_constraint_expression_evaluation import format_constraint_evaluation
from ahbicht.expressions.requirement_constraint_expression_evaluation import requirement_constraint_evaluation


# pylint: disable=no-self-use, invalid-name
class AhbExpressionTransformer(Transformer):
    """
    Transformer, that evaluates the trees built from the ahb expressions.
    The input are the conditions as defined in the AHBs in the form of ConditionNodes.
    It returns a list of the seperated requirement indicators with
    their respective condition expressions already evaluated to booleans.
    """

    def __init__(self, entered_input: str):
        """
        The input are the evaluated format constraint conditions in the form of ConditionNodes.
        :param entered_input: dict(condition_keys, ConditionNode)
        """
        super().__init__()
        self.entered_input = entered_input

    def CONDITION_EXPRESSION(self, condition_expression: Token) -> str:
        """Returns the condition expression."""
        return condition_expression.value

    def PREFIX_OPERATOR(self, prefix_operator: Token) -> str:
        """Returns the prefix operator."""
        return prefix_operator.value

    def MODAL_MARK(self, modal_mark: Token) -> str:
        """Returns the modal mark."""
        return modal_mark.value

    @v_args(inline=True)  # Children are provided as *args instead of a list argument
    def single_requirement_indicator_expression(
        self, requirement_indicator, condition_expression
    ) -> ConditionCheckResult:
        """
        Evaluates the condition expression of the respective requirement indicator expression
        and returns a list of the seperated requirement indicators with
        their results of the condition check.
        """
        requirement_constraint_evaluation_result: RequirementConstraintEvaluationResult = (
            requirement_constraint_evaluation(condition_expression)
        )
        format_constraint_evaluation_result: FormatConstraintEvaluationResult = format_constraint_evaluation(
            requirement_constraint_evaluation_result.format_constraints_expression, self.entered_input
        )

        result_of_condition_check: ConditionCheckResult = ConditionCheckResult(
            requirement_indicator=requirement_indicator,
            requirement_constraint_evaluation_result=requirement_constraint_evaluation_result,
            format_constraint_evaluation_result=format_constraint_evaluation_result,
        )

        return result_of_condition_check

    @v_args(inline=True)  # Children are provided as *args instead of a list argument
    def requirement_indicator(self, requirement_indicator) -> ConditionCheckResult:
        """
        If there is no condition expression but only a requirement indicator,
        all evaluations are automatically set to True.
        """
        return ConditionCheckResult(
            requirement_indicator=requirement_indicator,
            requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                requirement_constraints_fulfilled=True,
                requirement_is_conditional=False,
                format_constraints_expression=None,
                hints=None,
            ),
            format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                format_constraints_fulfilled=True, error_message=None
            ),
        )

    # pylint: disable=(line-too-long)
    def ahb_expression(
        self, list_of_single_requirement_indicator_expressions: List[ConditionCheckResult]
    ) -> ConditionCheckResult:
        """
        Returns the requirement indicator with its condition expressions already evaluated to booleans.
        If there are more than one modal mark the first whose conditions are fulfilled is returned or the
        last of the list if they are all unfulfilled.
        """
        for single_requirement_indicator_expression in list_of_single_requirement_indicator_expressions:
            if (
                single_requirement_indicator_expression.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            ):
                # if there are more than one modal mark, the requirement is conditional
                # even if the one of modal mark itself is not, e.g. `Muss[1] Kann`
                if len(list_of_single_requirement_indicator_expressions) > 1:
                    single_requirement_indicator_expression.requirement_constraint_evaluation_result.requirement_is_conditional = (
                        True
                    )
                return single_requirement_indicator_expression
        return list_of_single_requirement_indicator_expressions[-1]


def evaluate_ahb_expression_tree(parsed_tree: Tree, entered_input: str) -> ConditionCheckResult:
    """
    Evaluates the tree built from the ahb expressions with the help of the AhbExpressionTransformer.

    :param parsed_tree: Tree
    :param entered_input: the conditions as defined in the AHBs in the form of ConditionNodes
    :return: the result of the overall condition check (including requirement constraints, format constraints,
        several modal marks)
    """
    try:
        result = AhbExpressionTransformer(entered_input).transform(parsed_tree)
    except VisitError as visit_err:
        raise visit_err.orig_exc

    return result



"""
This module contains the classes for the evaluation results.
A "result" is the outcome of a evaluation. It requires actual data to be present.
"""

from typing import Optional

import attr

# pylint: disable=too-few-public-methods, no-member, no-self-use, unused-argument
from marshmallow import Schema, fields, post_load


@attr.s(auto_attribs=True, kw_only=True)
class RequirementConstraintEvaluationResult:
    """
    A class for the result of the requirement constraint evaluation.
    """

    requirement_constraints_fulfilled: bool  # true if condition expression in regard to
    # requirement constraints evaluates to true
    requirement_is_conditional: bool  # true if it is dependent on requirement constraints
    format_constraints_expression: Optional[str]
    hints: Optional[
        str
    ]  # Hint text that should be displayed in the frontend, e.g. "[501] Hinweis: 'ID der Messlokation'"


class RequirementConstraintEvaluationResultSchema(Schema):
    """
    A schema to (de-)serialize RequirementConstraintEvaluationResult
    """

    requirement_constraints_fulfilled = fields.Boolean()
    requirement_is_conditional = fields.Boolean()
    format_constraints_expression = fields.String()
    hints = fields.String()

    @post_load
    def deserialize(self, data, **kwargs) -> RequirementConstraintEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual RequirementConstraintEvaluationResult
        :param data:
        :param kwargs:
        :return:
        """
        return RequirementConstraintEvaluationResult(**data)


@attr.s(auto_attribs=True, kw_only=True)
class FormatConstraintEvaluationResult:
    """
    A class for the result of the format constraint evaluation.
    """

    format_constraints_fulfilled: bool  # true if data entered obey the format constraint expression
    error_message: Optional[str]  # All error messages that lead to not fulfilling the format constraint expression


class FormatConstraintEvaluationResultSchema(Schema):
    """
    A class to (de-)serialize FormatConstraintEvaluationResult
    """

    format_constraints_fulfilled = fields.Boolean()
    error_message = fields.String(allow_none=True)

    @post_load
    def deserialize(self, data, **kwargs) -> FormatConstraintEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual FormatConstraintEvaluationResult
        :param data:
        :param kwargs:
        :return:
        """
        return FormatConstraintEvaluationResult(**data)


@attr.s(auto_attribs=True, kw_only=True)
class ConditionCheckResult:
    """
    A class for the result of the overall condition check.
    """

    requirement_indicator: str  # i.e. "Muss", "Soll", "Kann", "X", "O", "U"
    requirement_constraint_evaluation_result: RequirementConstraintEvaluationResult
    format_constraint_evaluation_result: FormatConstraintEvaluationResult


class ConditionCheckResultSchema(Schema):
    """
    A schema to (de-)serialize ConditionCheckResults
    """

    requirement_indicator = fields.String(required=True, allow_none=True)
    requirement_constraint_evaluation_result = fields.Nested(RequirementConstraintEvaluationResultSchema())
    format_constraint_evaluation_result = fields.Nested(FormatConstraintEvaluationResultSchema())

    @post_load
    def deserialize(self, data, **kwargs) -> ConditionCheckResult:
        """
        Converts the barely typed data dictionary into an actual ConditionCheckResult
        :param data:
        :param kwargs:
        :return:
        """
        return ConditionCheckResult(**data)



"""
Tests that the parsed trees are JSON serializable
"""
import json
import uuid
from typing import TypeVar

import pytest
from lark import Token, Tree
from marshmallow import Schema, ValidationError

from ahbicht.condition_check_results import (
    ConditionCheckResult,
    ConditionCheckResultSchema,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedFormatConstraint,
    EvaluatedFormatConstraintSchema,
)
from ahbicht.json_serialization.tree_schema import TreeSchema

T = TypeVar("T")


def _test_serialization_roundtrip(serializable_object: T, schema: Schema, expected_json_dict: dict) -> T:
    """
    Serializes the serializable_object using the provided schema,
    asserts, that the result is equal to the expected_json_dict
    then deserializes it again and asserts on equality with the original serializable_object
    :returns the deserialized_object
    """
    json_string = schema.dumps(serializable_object)
    assert json_string is not None
    actual_json_dict = json.loads(json_string)
    assert actual_json_dict == expected_json_dict
    deserialized_object = schema.loads(json_data=json_string)
    assert isinstance(deserialized_object, type(serializable_object))
    assert deserialized_object == serializable_object
    return deserialized_object


class TestJsonSerialization:
    @pytest.mark.parametrize(
        "tree, expected_json_dict",
        [
            pytest.param(
                Tree(
                    "or_composition",
                    [
                        Tree("condition_key", [Token("INT", "53")]),
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"type": "INT", "value": "53"}, "tree": None}],
                                "type": "condition_key",
                            },
                        },
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "1"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "2"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                ],
                                "type": "and_composition",
                            },
                        },
                    ],
                    "type": "or_composition",
                },
            )
        ],
    )
    def test_tree_serialization(self, tree: Tree, expected_json_dict: dict):
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "condition_string, expected_json_dict",
        [
            pytest.param(
                "Muss [2] U ([3] O [4])[901] U [555]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {"token": {"type": "MODAL_MARK", "value": "Muss"}, "tree": None},
                                    {
                                        "token": {
                                            "type": "CONDITION_EXPRESSION",
                                            "value": " [2] U ([3] O [4])[901] U [555]",
                                        },
                                        "tree": None,
                                    },
                                ],
                                "type": "single_requirement_indicator_expression",
                            },
                        }
                    ],
                    "type": "ahb_expression",
                },
            )
        ],
    )
    def test_single_requirement_indicator_expression_serialization(
        self, condition_string: str, expected_json_dict: dict
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(condition_string)
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "evaluated_format_constraint, expected_json_dict",
        [
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                {"format_constraint_fulfilled": True, "error_message": None},
            ),
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something is wrong"),
                {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
            ),
        ],
    )
    def test_evaluated_format_constraint_serialization(
        self, evaluated_format_constraint: EvaluatedFormatConstraint, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            evaluated_format_constraint, EvaluatedFormatConstraintSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "invalid_content_evaluation_result_dict",
        [
            pytest.param({}, id="empty dict"),
            pytest.param({"format_constraints": {}, "hints": {}}, id="missing requirement constraints"),
            pytest.param({"requirement_constraints": {}, "hints": {}}, id="missing format_constraints"),
        ],
    )
    def test_validation_errors_on_content_evaluation_result_deserialization(
        self, invalid_content_evaluation_result_dict: dict
    ):
        schema = ContentEvaluationResultSchema()
        with pytest.raises(ValidationError):
            schema.load(invalid_content_evaluation_result_dict)

    @pytest.mark.parametrize(
        "content_evaluation_result, expected_json_dict",
        [
            pytest.param(
                ContentEvaluationResult(
                    hints={"501": "foo", "502": "bar", "503": None},
                    format_constraints={
                        "901": EvaluatedFormatConstraint(
                            format_constraint_fulfilled=False, error_message="something is wrong"
                        ),
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "1": ConditionFulfilledValue.NEUTRAL,
                        "2": ConditionFulfilledValue.FULFILLED,
                        "3": ConditionFulfilledValue.UNFULFILLED,
                        "4": ConditionFulfilledValue.UNKNOWN,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                ),
                {
                    "hints": {"501": "foo", "502": "bar", "503": None},
                    "format_constraints": {
                        "902": {"format_constraint_fulfilled": True, "error_message": None},
                        "901": {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
                    },
                    "requirement_constraints": {"1": "NEUTRAL", "2": "FULFILLED", "3": "UNFULFILLED", "4": "UNKNOWN"},
                    "id": "d106f335-f663-4d14-9636-4f43a883ad26",
                },
            ),
        ],
    )
    def test_content_evaluation_result_serialization(
        self, content_evaluation_result: ContentEvaluationResult, expected_json_dict: dict
    ):
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        deserialized_object = _test_serialization_roundtrip(
            content_evaluation_result, ContentEvaluationResultSchema(), expected_json_dict
        )
        for rc_evaluation_result in deserialized_object.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)

    @pytest.mark.parametrize(
        "condition_check_result, expected_json_dict",
        [
            pytest.param(
                ConditionCheckResult(
                    requirement_indicator="Muss",
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        error_message="hello", format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        hints="foo bar",
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                        format_constraints_expression="[asd]",
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": "hello",
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": "[asd]",
                        "hints": "foo bar",
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "Muss",
                },
            ),
        ],
    )
    def test_condition_check_result_serialization(
        self, condition_check_result: ConditionCheckResult, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(condition_check_result, ConditionCheckResultSchema(), expected_json_dict)



"""
Requirement Constraint (RC) Evaluators are evaluators that check if data are required under given circumstances.
Typical usecases are for example
* you must only provide a Gerätenummer if the Transaktionsgrund is f.e. 'E08'
* you must only provide an Ausbaudatum if the meter is being removed f.e. 'Z02'
"""
import asyncio
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluationContext
from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue


# pylint: disable=no-self-use
class RcEvaluator(Evaluator, ABC):
    """
    Base class of all Requirement Constraint (RC) evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    def __init__(self, evaluatable_data: EvaluatableData):
        """
        Instantiate any evaluator by providing some evaluatable data from a message.
        :param evaluatable_data:
        """
        if not evaluatable_data:
            raise ValueError("Evaluatable data have to be provided to any evaluator.")
        self.evaluatable_data = evaluatable_data

    @abstractmethod
    def _get_default_context(self) -> EvaluationContext:
        raise NotImplementedError("Has to be implemented in inheriting class")

    async def evaluate_single_condition(
        self, condition_key: str, context: Optional[EvaluationContext] = None
    ) -> ConditionFulfilledValue:
        """
        Evaluates the condition with the given key.
        :param context: optional content_evaluation context, if not the default context will be used
        :param condition_key: key of the condition, e.g. "78"
        :return:
        """
        evaluation_method = self.get_evaluation_method(condition_key)
        if evaluation_method is None:
            raise NotImplementedError(f"There is no content_evaluation method for condition '{condition_key}'")
        if context is None:
            context = self._get_default_context()
        result = await evaluation_method(context)
        return result

    async def evaluate_conditions(
        self, condition_keys: List[str], condition_keys_with_context: Optional[Dict[str, EvaluationContext]] = None
    ) -> Dict[str, ConditionFulfilledValue]:
        """
        Validate all the conditions provided in condition_keys in their respective context.
        """
        if condition_keys_with_context is None:
            tasks = [self.evaluate_single_condition(condition_key, None) for condition_key in condition_keys]
        else:
            tasks = []
            for condition_key in condition_keys:
                if condition_key not in condition_keys_with_context:
                    tasks.append(self.evaluate_single_condition(condition_key, None))
                else:
                    tasks.append(
                        self.evaluate_single_condition(condition_key, condition_keys_with_context[condition_key])
                    )

        results = await asyncio.gather(*tasks)

        result = dict(zip(condition_keys, results))
        return result


class DictBasedRcEvaluator(RcEvaluator):
    """
    A requirement constraint evaluator that is initialized with a prefilled dictionary.
    """

    def __init__(self, results: Dict[str, ConditionFulfilledValue]):
        """
        Initialize with a dictionary that contains all the requirement constraint evaluation results.
        :param results:
        """
        super().__init__(evaluatable_data=EvaluatableData(edifact_seed=results))
        self._results: Dict[str, ConditionFulfilledValue] = results

    def _get_default_context(self) -> EvaluationContext:
        raise NotImplementedError()

    # pylint:disable=unused-argument
    async def evaluate_single_condition(
        self, condition_key: str, context: Optional[EvaluationContext] = None
    ) -> ConditionFulfilledValue:
        try:
            return self._results[condition_key]
        except KeyError as key_error:
            raise NotImplementedError(f"No result was provided for condition '{condition_key}'.") from key_error



"""
Tests the expression builder module.
"""
import pytest

from ahbicht.expressions.condition_nodes import Hint, UnevaluatedFormatConstraint
from ahbicht.expressions.expression_builder import FormatConstraintExpressionBuilder, HintExpressionBuilder


class TestFormatConstraintExpressionBuilder:
    """
    Tests the format constraint expression builder.
    """

    def test_simple_construction(self):
        fc_1 = UnevaluatedFormatConstraint(condition_key="1")
        fc_2 = UnevaluatedFormatConstraint(condition_key="2")
        fc_3 = UnevaluatedFormatConstraint(condition_key="3")
        fc_4 = UnevaluatedFormatConstraint(condition_key="4")
        fceb = FormatConstraintExpressionBuilder(fc_1).lor(fc_2).land(fc_3).xor(fc_4)
        assert fceb.get_expression() == "(([1] O [2]) U [3]) X [4]"

        fceb_from_str = FormatConstraintExpressionBuilder("[7] U [8] U [9]").lor(fc_2).land("[3]").xor(fc_4)
        assert fceb_from_str.get_expression() == "((([7] U [8] U [9]) O [2]) U [3]) X [4]"


class TestHintExpressionBuilder:
    """
    Tests the hint expression builder.
    """

    @pytest.mark.parametrize(
        "init, other, expected",
        [
            pytest.param(Hint(condition_key="0", hint="foo"), None, "foo"),
            pytest.param(None, Hint(condition_key="0", hint="bar"), "bar"),
            pytest.param(Hint(condition_key="0", hint="foo"), Hint(condition_key="1", hint="bar"), "foo und bar"),
            pytest.param("foo", Hint(condition_key="1", hint="bar"), "foo und bar"),
            pytest.param("foo", "bar", "foo und bar"),
            pytest.param(None, None, None),
        ],
    )
    def test_logical_and(self, init, other, expected: Hint):
        builder = HintExpressionBuilder(init)
        actual = builder.land(other).get_expression()
        assert actual == expected

    @pytest.mark.parametrize(
        "init, other, expected",
        [
            pytest.param(Hint(condition_key="0", hint="foo"), None, "foo"),
            pytest.param(None, Hint(condition_key="0", hint="bar"), "bar"),
            pytest.param(Hint(condition_key="0", hint="foo"), Hint(condition_key="1", hint="bar"), "foo oder bar"),
            pytest.param("foo", Hint(condition_key="1", hint="bar"), "foo oder bar"),
            pytest.param("foo", "bar", "foo oder bar"),
            pytest.param(None, None, None),
        ],
    )
    def test_logical_or(self, init, other, expected: Hint):
        builder = HintExpressionBuilder(init)
        actual = builder.lor(other).get_expression()
        assert actual == expected

    @pytest.mark.parametrize(
        "init, other, expected",
        [
            pytest.param(Hint(condition_key="0", hint="foo"), None, "foo"),
            pytest.param(None, Hint(condition_key="0", hint="bar"), "bar"),
            pytest.param(
                Hint(condition_key="0", hint="foo"), Hint(condition_key="1", hint="bar"), "Entweder (foo) oder (bar)"
            ),
            pytest.param("foo", Hint(condition_key="1", hint="bar"), "Entweder (foo) oder (bar)"),
            pytest.param(
                "foo und x und y", Hint(condition_key="1", hint="bar"), "Entweder (foo und x und y) oder (bar)"
            ),
            pytest.param("foo", "bar", "Entweder (foo) oder (bar)"),
            pytest.param(None, None, None),
        ],
    )
    def test_logical_xor(self, init, other, expected: Hint):
        builder = HintExpressionBuilder(init)
        actual = builder.xor(other).get_expression()
        assert actual == expected

    def test_a_longer_concatenation(self):
        builder = HintExpressionBuilder("foo").land("bar").lor("asd").xor("xyz")
        assert builder.get_expression() == "Entweder (foo und bar oder asd) oder (xyz)"
